@@ -26,9 +26,7 @@
 #include <iconv.h>
 #include <ctype.h>
 
-#if defined(ENABLE_ICU)
-#include <unicode/unorm.h>
-#endif
+#include <xkbcommon/xkbcommon-compose.h>
 
 #include "fcitx/ime.h"
 #include "fcitx/instance.h"
@@ -47,7 +45,6 @@
 #if defined(ENABLE_LIBXML2)
 #include "isocodes.h"
 #endif
-#include "fcitx-compose-data.h"
 
 #define INVALID_COMPOSE_RESULT 0xffffffff
 
@@ -75,70 +72,8 @@ static void SaveKeyboardConfig(FcitxKeyboardConfig* fs);
 static INPUT_RETURN_VALUE FcitxKeyboardHotkeyToggleWordHint(void* arg);
 static inline boolean IsValidSym(FcitxKeySym keysym, unsigned int state);
 static inline boolean IsValidChar(uint32_t c);
-static uint32_t checkCompactTable(FcitxKeyboardLayout* layout, const FcitxComposeTableCompact *table);
-static uint32_t checkAlgorithmically(FcitxKeyboardLayout* layout);
 static uint32_t processCompose(FcitxKeyboardLayout* layout, uint32_t keyval, uint32_t state);
 
-static int
-compare_seq_index(const void *key, const void *value)
-{
-    const uint32_t *keysyms = (const uint32_t *)key;
-    const uint32_t *seq = (const uint32_t *)value;
-
-    if (keysyms[0] < seq[0])
-        return -1;
-    else if (keysyms[0] > seq[0])
-        return 1;
-    return 0;
-}
-
-static int
-compare_seq(const void *key, const void *value)
-{
-    int i = 0;
-    const uint32_t *keysyms = (const uint32_t *)key;
-    const uint32_t *seq = (const uint32_t *)value;
-
-    while (keysyms[i]) {
-        if (keysyms[i] < seq[i])
-            return -1;
-        else if (keysyms[i] > seq[i])
-            return 1;
-        i++;
-    }
-
-    return 0;
-}
-
-static const uint32_t fcitx_compose_ignore[] = {
-    FcitxKey_Shift_L,
-    FcitxKey_Shift_R,
-    FcitxKey_Control_L,
-    FcitxKey_Control_R,
-    FcitxKey_Caps_Lock,
-    FcitxKey_Shift_Lock,
-    FcitxKey_Meta_L,
-    FcitxKey_Meta_R,
-    FcitxKey_Alt_L,
-    FcitxKey_Alt_R,
-    FcitxKey_Super_L,
-    FcitxKey_Super_R,
-    FcitxKey_Hyper_L,
-    FcitxKey_Hyper_R,
-    FcitxKey_Mode_switch,
-    FcitxKey_ISO_Level3_Shift,
-    FcitxKey_VoidSymbol
-};
-
-static inline
-void Ucs4ToUtf8(iconv_t conv, uint32_t ucs4, char* utf8)
-{
-    size_t inbytes = sizeof(uint32_t);
-    size_t outbytes = UTF8_MAX_LENGTH;
-    IconvStr in = (IconvStr) &ucs4;
-    iconv(conv, &in, &inbytes,  &utf8, &outbytes);
-}
-
 CONFIG_DESC_DEFINE(GetKeyboardConfigDesc, "fcitx-keyboard.desc")
 
 FCITX_DEFINE_PLUGIN(fcitx_keyboard, ime2, FcitxIMClass2) = {
@@ -357,12 +292,29 @@ static const char* FindBestLanguage(FcitxIsoCodes* isocodes, const char* hint, U
 void* FcitxKeyboardCreate(FcitxInstance* instance)
 {
     FcitxKeyboard* keyboard = fcitx_utils_malloc0(sizeof(FcitxKeyboard));
-    keyboard->owner = instance;
     if (!LoadKeyboardConfig(keyboard, &keyboard->config))
     {
         free(keyboard);
         return NULL;
     }
+    const char* locale = getenv("LC_ALL");
+    if (!locale)
+        locale = getenv("LC_CTYPE");
+    if (!locale)
+        locale = getenv("LANG");
+    if (!locale)
+        locale = "C";
+
+    keyboard->owner = instance;
+    keyboard->xkbContext = xkb_context_new(XKB_CONTEXT_NO_FLAGS);
+
+    if (keyboard->xkbContext) {
+        xkb_context_set_log_level(keyboard->xkbContext, XKB_LOG_LEVEL_CRITICAL);
+    }
+
+    keyboard->xkbComposeTable = keyboard->xkbContext ? xkb_compose_table_new_from_locale(keyboard->xkbContext, locale, XKB_COMPOSE_COMPILE_NO_FLAGS) : NULL;
+    keyboard->xkbComposeState = keyboard->xkbComposeTable ? xkb_compose_state_new(keyboard->xkbComposeTable, XKB_COMPOSE_STATE_NO_FLAGS) : NULL;
+
     char* localepath = fcitx_utils_get_fcitx_path("localedir");
     bindtextdomain("xkeyboard-config", localepath);
     bind_textdomain_codeset("xkeyboard-config", "UTF-8");
@@ -578,8 +530,9 @@ void  FcitxKeyboardResetIM(void *arg)
     keyboard->cursor_moved = false;
     keyboard->buffer[0][0] = '\0';
     keyboard->cursorPos = 0;
-    keyboard->composeBuffer[0] = 0;
-    keyboard->n_compose = 0;
+    if (keyboard->xkbComposeState) {
+        xkb_compose_state_reset(keyboard->xkbComposeState);
+    }
 }
 
 void FcitxKeyboardOnClose(void* arg, FcitxIMCloseEventType event)
@@ -741,10 +694,11 @@ INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int
                  FcitxHotkeyIsHotKey(sym, state, FCITX_HYPHEN_APOS))) {
                 char buf[UTF8_MAX_LENGTH + 1];
                 memset(buf, 0, sizeof(buf));
-                if (result)
-                    Ucs4ToUtf8(keyboard->iconv, result, buf);
-                else
-                    Ucs4ToUtf8(keyboard->iconv, FcitxKeySymToUnicode(sym), buf);
+                if (result) {
+                    fcitx_ucs4_to_utf8(result, buf);
+                } else {
+                    fcitx_ucs4_to_utf8(FcitxKeySymToUnicode(sym), buf);
+                }
                 size_t charlen = strlen(buf);
 
                 size_t len = strlen(keyboard->buffer[0]);
@@ -804,7 +758,7 @@ INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int
             if (result) {
                 char buf[UTF8_MAX_LENGTH + 1];
                 memset(buf, 0, sizeof(buf));
-                Ucs4ToUtf8(keyboard->iconv, result, buf);
+                fcitx_ucs4_to_utf8(result, buf);
                 FcitxInstanceCommitString(instance, currentIC, buf);
                 irv = 0;
             }
@@ -817,7 +771,7 @@ INPUT_RETURN_VALUE FcitxKeyboardDoInput(void *arg, FcitxKeySym sym, unsigned int
     if (result) {
         char buf[UTF8_MAX_LENGTH + 1];
         memset(buf, 0, sizeof(buf));
-        Ucs4ToUtf8(keyboard->iconv, result, buf);
+        fcitx_ucs4_to_utf8(result, buf);
         FcitxInstanceCommitString(instance, currentIC, buf);
         return IRV_FLAG_UPDATE_INPUT_WINDOW | IRV_FLAG_RESET_INPUT;
     }
@@ -2777,196 +2731,34 @@ boolean IsValidChar(uint32_t c) {
 uint32_t
 processCompose(FcitxKeyboardLayout* layout, uint32_t keyval, uint32_t state)
 {
-    int i;
     FcitxKeyboard* keyboard = layout->owner;
-
-    for (i = 0; fcitx_compose_ignore[i] != FcitxKey_VoidSymbol; i++) {
-        if (keyval == fcitx_compose_ignore[i])
-            return 0;
-    }
-
-    keyboard->composeBuffer[keyboard->n_compose ++] = keyval;
-    keyboard->composeBuffer[keyboard->n_compose] = 0;
-
-    uint32_t result;
-    if ((result = checkCompactTable(layout, &fcitx_compose_table_compact))) {
-        // qDebug () << "checkCompactTable ->true";
-        return result;
-    }
-
-    if ((result = checkAlgorithmically(layout))) {
-        // qDebug () << "checkAlgorithmically ->true";
-        return result;
-    }
-
-    if (keyboard->n_compose > 1) {
-        keyboard->composeBuffer[0] = 0;
-        keyboard->n_compose = 0;
-        return INVALID_COMPOSE_RESULT;
-    } else {
-        keyboard->composeBuffer[0] = 0;
-        keyboard->n_compose = 0;
+    if (!keyboard->xkbComposeState) {
         return 0;
     }
+
+    enum xkb_compose_feed_result result = xkb_compose_state_feed(keyboard->xkbComposeState, keyval);
+    if (result == XKB_COMPOSE_FEED_IGNORED) {
+        return 0;
+    }
+
+    enum xkb_compose_status status = xkb_compose_state_get_status(keyboard->xkbComposeState);
+    if (status == XKB_COMPOSE_NOTHING) {
+        return 0;
+    } else if (status == XKB_COMPOSE_COMPOSED) {
+        char buffer[UTF8_MAX_LENGTH + 1] = {'\0', '\0', '\0', '\0', '\0', '\0', '\0'};
+        int length = xkb_compose_state_get_utf8(keyboard->xkbComposeState, buffer, sizeof(buffer));
+        xkb_compose_state_reset(keyboard->xkbComposeState);
+        if (length == 0) {
+            return INVALID_COMPOSE_RESULT;
+        }
+
+        uint32_t c = 0;
+        fcitx_utf8_get_char(buffer, &c);
+        return c;
+    } else if (status == XKB_COMPOSE_CANCELLED) {
+        xkb_compose_state_reset(keyboard->xkbComposeState);
+    }
+
+    return INVALID_COMPOSE_RESULT;
 }
 
-uint32_t
-checkCompactTable(FcitxKeyboardLayout* layout, const FcitxComposeTableCompact *table)
-{
-    int row_stride;
-    const uint32_t *seq_index;
-    const uint32_t *seq;
-    int i;
-    FcitxKeyboard* keyboard = layout->owner;
-
-    /* Will never match, if the sequence in the compose buffer is longer
-    * than the sequences in the table. Further, compare_seq (key, val)
-    * will overrun val if key is longer than val. */
-    if (keyboard->n_compose > table->max_seq_len)
-        return 0;
-
-    seq_index = (const uint32_t *)bsearch(keyboard->composeBuffer,
-                                         table->data, table->n_index_size,
-                                         sizeof(uint32_t) * table->n_index_stride,
-                                         compare_seq_index);
-
-    if (!seq_index) {
-        return 0;
-    }
-
-    if (seq_index && keyboard->n_compose == 1) {
-        return INVALID_COMPOSE_RESULT;
-    }
-
-    seq = NULL;
-    for (i = keyboard->n_compose - 1; i < table->max_seq_len; i++) {
-        row_stride = i + 1;
-
-        if (seq_index[i + 1] - seq_index[i] > 0) {
-            seq = (const uint32_t *) bsearch(keyboard->composeBuffer + 1,
-                                            table->data + seq_index[i], (seq_index[i + 1] - seq_index[i]) / row_stride,
-                                            sizeof(uint32_t) * row_stride,
-                                            compare_seq);
-            if (seq) {
-                if (i == keyboard->n_compose - 1)
-                    break;
-                else {
-                    return INVALID_COMPOSE_RESULT;
-                }
-            }
-        }
-    }
-
-    if (!seq) {
-        return 0;
-    } else {
-        uint32_t value;
-        value = seq[row_stride - 1];
-        keyboard->composeBuffer[0] = 0;
-        keyboard->n_compose = 0;
-        return value;
-    }
-    return 0;
-}
-
-#define IS_DEAD_KEY(k) \
-    ((k) >= FcitxKey_dead_grave && (k) <= (FcitxKey_dead_dasia+1))
-
-uint32_t
-checkAlgorithmically(FcitxKeyboardLayout* layout)
-{
-#if defined(ENABLE_ICU)
-    int i;
-    UChar combination_buffer[FCITX_MAX_COMPOSE_LEN];
-    FcitxKeyboard* keyboard = layout->owner;
-
-    if (keyboard->n_compose >= FCITX_MAX_COMPOSE_LEN)
-        return 0;
-
-    for (i = 0;i < keyboard->n_compose &&
-             IS_DEAD_KEY(keyboard->composeBuffer[i]);i++) {
-    }
-    if (i == keyboard->n_compose)
-        return INVALID_COMPOSE_RESULT;
-
-    if (i > 0 && i == keyboard->n_compose - 1) {
-        combination_buffer[0] = FcitxKeySymToUnicode((FcitxKeySym) keyboard->composeBuffer[i]);
-        combination_buffer[keyboard->n_compose] = 0;
-        i--;
-        while (i >= 0) {
-            switch (keyboard->composeBuffer[i]) {
-#define CASE(keysym, unicode) \
-case FcitxKey_dead_##keysym: combination_buffer[i + 1] = unicode; break
-            CASE(grave, 0x0300);
-            CASE(acute, 0x0301);
-            CASE(circumflex, 0x0302);
-            CASE(tilde, 0x0303);   /* Also used with perispomeni, 0x342. */
-            CASE(macron, 0x0304);
-            CASE(breve, 0x0306);
-            CASE(abovedot, 0x0307);
-            CASE(diaeresis, 0x0308);
-            CASE(hook, 0x0309);
-            CASE(abovering, 0x030A);
-            CASE(doubleacute, 0x030B);
-            CASE(caron, 0x030C);
-            CASE(abovecomma, 0x0313);         /* Equivalent to psili */
-            CASE(abovereversedcomma, 0x0314); /* Equivalent to dasia */
-            CASE(horn, 0x031B);    /* Legacy use for psili, 0x313 (or 0x343). */
-            CASE(belowdot, 0x0323);
-            CASE(cedilla, 0x0327);
-            CASE(ogonek, 0x0328);  /* Legacy use for dasia, 0x314.*/
-            CASE(iota, 0x0345);
-            CASE(voiced_sound, 0x3099);    /* Per Markus Kuhn keysyms.txt file. */
-            CASE(semivoiced_sound, 0x309A);    /* Per Markus Kuhn keysyms.txt file. */
-
-            /* The following cases are to be removed once xkeyboard-config,
-            * xorg are fully updated.
-            */
-                /* Workaround for typo in 1.4.x xserver-xorg */
-            case 0xfe66: combination_buffer[i+1] = 0x314; break;
-            /* CASE(dasia, 0x314); */
-            /* CASE(perispomeni, 0x342); */
-            /* CASE(psili, 0x343); */
-#undef CASE
-            default:
-                combination_buffer[i + 1] = FcitxKeySymToUnicode((FcitxKeySym) keyboard->composeBuffer[i]);
-            }
-            i--;
-        }
-
-        /* If the buffer normalizes to a single character,
-        * then modify the order of combination_buffer accordingly, if necessary,
-        * and return TRUE.
-        **/
-#if 0
-        if (check_normalize_nfc(combination_buffer, keyboard->n_compose)) {
-            gunichar value;
-            combination_utf8 = g_ucs4_to_utf8(combination_buffer, -1, NULL, NULL, NULL);
-            nfc = g_utf8_normalize(combination_utf8, -1, G_NORMALIZE_NFC);
-
-            value = g_utf8_get_char(nfc);
-            gtk_im_context_simple_commit_char(GTK_IM_CONTEXT(context_simple), value);
-            context_simple->compose_buffer[0] = 0;
-
-            g_free(combination_utf8);
-            g_free(nfc);
-
-            return TRUE;
-        }
-#endif
-        UErrorCode state = U_ZERO_ERROR;
-        UChar result[FCITX_MAX_COMPOSE_LEN + 1];
-        i = unorm_normalize(combination_buffer, keyboard->n_compose, UNORM_NFC, 0, result, FCITX_MAX_COMPOSE_LEN + 1, &state);
-
-        // qDebug () << "combination_buffer = " << QString::fromUtf16(combination_buffer) << "keyboard->n_compose" << keyboard->n_compose;
-        // qDebug () << "result = " << QString::fromUtf16(result) << "i = " << i << state;
-
-        if (i == 1) {
-            keyboard->composeBuffer[0] = 0;
-            keyboard->n_compose = 0;
-            return result[0];
-        }
-    }
-#endif
-    return 0;
-}
